modes = [
    'normal',
    'multiply',
    'screen',
    'overlay',
    'darken',
    'lighten',
    'colorDodge',
    'colorBurn',
    'softLight',
    'hardLight',
    'difference',
    'exclusion',
    'hue',
    'saturation',
    'color',
    'luminosity',
    # 'clear',
    # 'copy',
    # 'sourceIn',
    # 'sourceOut',
    # 'sourceAtop',
    # 'destinationOver',
    # 'destinationIn',
    # 'destinationOut',
    # 'destinationAtop',
    # 'xOR',
    # 'plusDarker',
    # 'plusLighter',
]

cellSize = 100
size(cellSize, cellSize * len(modes))
fill(0.5)
rect(0, 0, width(), height())

for mode in modes:
    blendMode(mode)

    fill(0, 0.4, 1)
    rect(0, 0, 0.75 * cellSize, 0.75 * cellSize)
    fill(1, 0, 0.4)
    rect(0.25 * cellSize, 0.25 * cellSize, 0.75 * cellSize, 0.75 * cellSize)
    translate(0, cellSize)



size(600, 120)

proofText = 'вгджзийклнпттттфцчшщъьюѝ'
font('../fonts/SourceSerifPro-Regular.otf')
fontSize(34)

w, _ = textSize(proofText)
assert round(w, 1) == 498.0
text(proofText, (12, 74))

language("bg")
w, _ = textSize(proofText)
assert round(w, 1) == 537.7
text(proofText, (12, 33))



import operator

path1 = BezierPath()
path1.rect(25, 25, 100, 100)
path2 = BezierPath()
path2.oval(75, 75, 100, 100)

operations = [
    (path1, path2, operator.or_),
    (path1, path2, operator.and_),
    (path1, path2, operator.mod),
    (path2, path1, operator.mod),
    (path1, path2, operator.xor),
]

size(200, 200 * len(operations))
fill(1)
rect(0, 0, width(), height())

stroke(0)
strokeWidth(4)
fill(0.9)

for path1, path2, op in operations:
    path = op(path1, path2)
    drawPath(path)
    translate(0, 200)



from drawbot_skia.path import BezierPath


def test_path_bounds():
    path = BezierPath()
    assert path.bounds() is None
    path.rect(10, 20, 30, 40)
    assert path.bounds() == (10, 20, 40, 60)


def test_path_controlPointBounds():
    path = BezierPath()
    assert path.controlPointBounds() is None
    path.moveTo((0, 0))
    path.curveTo((50, 100), (100, 100), (150, 0))
    assert path.bounds() == (0.0, 0.0, 150.0, 75.0)
    assert path.controlPointBounds() == (0.0, 0.0, 150.0, 100.0)


def test_path_copy():
    path1 = BezierPath()
    path1.rect(0, 0, 100, 100)
    path2 = path1.copy()
    path1.translate(50, 20)
    assert path1.bounds() == (50.0, 20.0, 150.0, 120.0)
    assert path2.bounds() == (0.0, 0.0, 100.0, 100.0)


def test_path_point_args():
    path1 = BezierPath()
    path1.moveTo([0, 0])
    path1.lineTo([0, 100])
    path1.curveTo([50, 100], [100, 100], [200, 0])


def test_path_line_args():
    path1 = BezierPath()
    path1.line([0, 0], [0, 100])




path1 = BezierPath()
path1.rect(25, 25, 100, 100)
path2 = BezierPath()
path2.oval(75, 75, 100, 100)

operations = [
    (path1, path2, "union"),
    (path1, path2, "intersection"),
    (path1, path2, "difference"),
    (path2, path1, "difference"),
    (path1, path2, "xor"),
]

size(200, 200 * (len(operations) + 1))
fill(1)
rect(0, 0, width(), height())

stroke(0)
strokeWidth(4)
fill(0.9)

for path1, path2, op in operations:
    path = getattr(path1, op)(path2)
    drawPath(path)
    translate(0, 200)

path = path1.copy()
path.appendPath(path2)
path.removeOverlap()
drawPath(path)



size(200, 400)

path = BezierPath()
path.polygon((80, 28), (50, 146), (146, 152), (172, 78))

with savedState():
    clipPath(path)
    scale(200/512)
    image("../images/drawbot.png", (0, 0))

translate(0, 200)

with savedState():
    scale(200/512)
    image("../images/drawbot.png", (0, 0))



from random import randint, random, seed

size(200, 100)

fill(1)
rect(0, 0, 200, 100)

fill(0)
path = BezierPath()
path.text("fiets", offset=(9, 18), font="../fonts/SourceSerifPro-Regular.otf", fontSize=95)

clipPath(path)

seed(8)

for i in range(79):
    fill(random(), random(), 0.5, random())
    oval(randint(0, 160), randint(0, 160), 40, 40)



import itertools
from fontTools.unicodedata import script
from unicodedata2 import category

# Monkeypatch bidi to use unicodedata2
import unicodedata2
import bidi.algorithm
bidi.algorithm.bidirectional = unicodedata2.bidirectional
bidi.algorithm.category = unicodedata2.category
bidi.algorithm.mirrored = unicodedata2.mirrored
from bidi.algorithm import (  # noqa: ignore E402
    get_empty_storage, get_base_level, get_embedding_levels,
    explicit_embed_and_overrides, resolve_weak_types,
    resolve_neutral_types, resolve_implicit_levels,
    reorder_resolved_levels, PARAGRAPH_LEVELS,
)
from bidi.mirror import MIRRORED  # noqa: ignore E402


UNKNOWN_SCRIPT = {"Zinh", "Zyyy", "Zxxx"}


def textSegments(txt):
    scripts = detectScript(txt)
    storage = getBiDiInfo(txt)

    levels = [None] * len(txt)
    for ch in storage['chars']:
        levels[ch['index']] = ch['level']

    prevLevel = storage['base_level']
    for i, level in enumerate(levels):
        if level is None:
            levels[i] = prevLevel
        else:
            prevLevel = level

    chars = list(zip(txt, scripts, levels))

    runLenghts = []
    for value, sub in itertools.groupby(chars, key=lambda item: item[1:]):
        runLenghts.append(len(list(sub)))

    segments = []
    index = 0
    for rl in runLenghts:
        nextIndex = index + rl
        segment = chars[index:nextIndex]
        runChars = "".join(ch for ch, script, bidiLevel in segment)
        _, script, bidiLevel = segment[0]
        segments.append((runChars, script, bidiLevel, index))
        index = nextIndex
    return segments, storage['base_level']


def reorderedSegments(segments, baseLevel):
    reorderedSegments = []
    isRTL = baseLevel % 2
    for value, sub in itertools.groupby(segments, key=lambda item: item[2] % 2):
        if isRTL == value:
            reorderedSegments.extend(sub)
        else:
            reorderedSegments.extend(reversed(list(sub)))
    if isRTL:
        reorderedSegments = list(reversed(reorderedSegments))
    assert len(reorderedSegments) == len(segments)
    return reorderedSegments


def detectScript(txt):
    charScript = [script(c) for c in txt]

    for i, ch in enumerate(txt):
        scr = charScript[i]
        if scr in UNKNOWN_SCRIPT:
            if i:
                scr = charScript[i-1]
            else:
                scr = None
            cat = category(ch)
            if ch in MIRRORED and cat == "Pe":
                scr = None
        charScript[i] = scr

    # Any unknowns should be mapped to the _next_ script
    prev = None
    for i in range(len(txt) - 1, -1, -1):
        if charScript[i] is None:
            charScript[i] = prev
        else:
            prev = charScript[i]

    # There may be unknowns at the end of the string, fall back to
    # preceding script
    prev = "Zxxx"  # last resort
    for i in range(len(txt)):
        if charScript[i] is None:
            charScript[i] = prev
        else:
            prev = charScript[i]

    assert None not in charScript

    return charScript


# copied from bidi/algorthm.py and modified to be more useful for us.

def getBiDiInfo(text, *, upper_is_rtl=False, base_dir=None, debug=False):
    """
    Set `upper_is_rtl` to True to treat upper case chars as strong 'R'
    for debugging (default: False).

    Set `base_dir` to 'L' or 'R' to override the calculated base_level.

    Set `debug` to True to display (using sys.stderr) the steps taken with the
    algorithm.

    Returns an info dict object and the display layout.
    """
    storage = get_empty_storage()

    if base_dir is None:
        base_level = get_base_level(text, upper_is_rtl)
    else:
        base_level = PARAGRAPH_LEVELS[base_dir]

    storage['base_level'] = base_level
    storage['base_dir'] = ('L', 'R')[base_level]

    get_embedding_levels(text, storage, upper_is_rtl, debug)
    assert len(text) == len(storage["chars"])
    for index, (ch, chInfo) in enumerate(zip(text, storage["chars"])):
        assert ch == chInfo["ch"]
        chInfo["index"] = index

    explicit_embed_and_overrides(storage, debug)
    resolve_weak_types(storage, debug)
    resolve_neutral_types(storage, debug)
    resolve_implicit_levels(storage, debug)
    reorder_resolved_levels(storage, debug)

    return storage



size(275, 400)

rect(0, 0, width(), height())
font("../fonts/MutatorSans.ttf")
fontSize(100)
fill(1)
text("ABCXYZ", (10, 15))
fontVariations(wght=600)
text("ABCXYZ", (10, 115))
variations = fontVariations(wdth=200)
assert {'wdth': 200, 'wght': 600.0} == variations, variations
text("ABCXYZ", (10, 215))
variations = fontVariations(wdth=1000, resetVariations=True)
assert {'wdth': 1000} == variations, variations
text("ABCXYZ", (10, 315))



import argparse
import sys
from .drawing import Drawing
from .runner import makeDrawbotNamespace, runScriptSource


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog="drawbot",
        description="Command line DrawBot tool.",
    )
    parser.add_argument(
        "drawbot_script",
        type=argparse.FileType("r"),
        help="The Drawbot script to run.",
    )
    parser.add_argument(
        "output_file",
        nargs="*",
        default=[],
        help="A filename for the output graphic.",
    )

    arguments = parser.parse_args(args)
    db = Drawing()
    namespace = makeDrawbotNamespace(db)
    runScriptSource(arguments.drawbot_script.read(), arguments.drawbot_script.name, namespace)
    for path in arguments.output_file:
        db.saveImage(path)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
import logging
import os
import pathlib
import tempfile
import skia


class Document(ABC):

    # pageWidth
    # pageHeight

    @property
    @abstractmethod
    def isDrawing(self):
        return ...

    @abstractmethod
    def beginPage(self, width: int, height: int) -> skia.Canvas:
        return ...

    @abstractmethod
    def endPage(self):
        ...

    @abstractmethod
    def setFrameDuration(self, duration):
        ...

    @abstractmethod
    def saveImage(self, path, **kwargs):
        ...


DEFAULT_FRAMEDURATION = 1/10


class RecordingDocument(Document):

    def __init__(self):
        self._pictures = []
        self._frameDurations = []
        self._currentRecorder = None
        self._currentFrameDuration = DEFAULT_FRAMEDURATION
        self.pageWidth = self.pageHeight = None

    @property
    def isDrawing(self):
        return self._currentRecorder is not None

    def beginPage(self, width, height):
        assert self._currentRecorder is None
        self.pageWidth = width
        self.pageHeight = height
        self._currentRecorder = skia.PictureRecorder()
        return self._currentRecorder.beginRecording(width, height)

    def endPage(self):
        self._pictures.append(self._currentRecorder.finishRecordingAsPicture())
        self._frameDurations.append(self._currentFrameDuration)
        self._currentRecorder = None
        self._currentFrameDuration = DEFAULT_FRAMEDURATION
        self.pageWidth = self.pageHeight = None

    def setFrameDuration(self, duration):
        self._currentFrameDuration = duration

    def saveImage(self, path, **kwargs):
        path = pathlib.Path(path).resolve()
        suffix = path.suffix.lower().lstrip(".")
        methodName = f"_saveImage_{suffix}"
        method = getattr(self, methodName, None)
        if method is None:
            raise ValueError(f"unsupported file type: {suffix}")
        method(path, **kwargs)

    def _saveImage_pdf(self, path, **kwargs):
        stream = skia.FILEWStream(os.fspath(path))
        with skia.PDF.MakeDocument(stream) as document:
            for picture in self._pictures:
                x, y, width, height = picture.cullRect()
                assert x == 0 and y == 0
                with document.page(width, height) as canvas:
                    canvas.drawPicture(picture)
        stream.flush()

    def _saveImage_svg(self, path, **kwargs):
        for picture, framePath in _iteratePictures(self._pictures, path):
            x, y, width, height = picture.cullRect()
            assert x == 0 and y == 0
            stream = skia.FILEWStream(os.fspath(framePath))
            canvas = skia.SVGCanvas.Make((width, height), stream)
            canvas.drawPicture(picture)
            del canvas
            stream.flush()

    def _saveImage_png(self, path, **kwargs):
        _savePixelImages(self._pictures, path, skia.kPNG)

    def _saveImage_jpeg(self, path, **kwargs):
        _savePixelImages(self._pictures, path, skia.kJPEG, whiteBackground=True)

    _saveImage_jpg = _saveImage_jpeg

    def _saveImage_mp4(self, path, codec="libx264", **kwargs):
        from .ffmpeg import generateMP4
        if not self._pictures:
            # Empty mp4?
            return
        frameRate = max(1, round(1 / self._frameDurations[-1]))
        if len(set(self._frameDurations)) != 1:
            logging.warning("ignoring varying frame durations for mp4 export")
        with tempfile.TemporaryDirectory(prefix="drawbot-skia-") as tempDir:
            tempDir = pathlib.Path(tempDir)
            imagePath = tempDir / "frame.png"
            _savePixelImages(
                self._pictures,
                imagePath,
                skia.kPNG,
                whiteBackground=True,
                singlePage=False,
            )
            imagesTemplate = os.fspath(tempDir / "frame_%d.png")
            generateMP4(imagesTemplate, path, frameRate, codec=codec)


def _savePixelImages(pictures, path, format, whiteBackground=False, singlePage=None):
    for picture, framePath in _iteratePictures(pictures, path, singlePage):
        _savePixelImage(picture, framePath, format, whiteBackground=whiteBackground)


def _iteratePictures(pictures, path, singlePage=None):
    if singlePage is None:
        singlePage = len(pictures) == 1
    for index, picture in enumerate(pictures):
        if singlePage:
            framePath = path
        else:
            framePath = path.parent / f"{path.stem}_{index}{path.suffix}"
        yield picture, framePath


def _savePixelImage(picture, path, format, whiteBackground=False):
    x, y, width, height = picture.cullRect()
    assert x == 0 and y == 0
    surface = skia.Surface(int(width), int(height))
    with surface as canvas:
        if whiteBackground:
            canvas.clear(skia.ColorWHITE)
        canvas.drawPicture(picture)
    image = surface.makeImageSnapshot()
    image.save(os.fspath(path), format)


class PixelDocument(Document):
    ...


class MP4Document(PixelDocument):
    ...


class PDFDocument(Document):
    ...


class SVGDocument(Document):
    ...



import os


def makeNamespace(*objects, **kwargs):
    namespace = kwargs
    for obj in objects:
        for name in dir(obj):
            if not name.startswith("_"):
                namespace[name] = getattr(obj, name)
    return namespace


def makeDrawbotNamespace(drawbot):
    from .path import BezierPath
    import math
    import random
    additionalNames = dict(BezierPath=BezierPath)
    for name in ["random", "randint", "choice", "shuffle"]:
        additionalNames[name] = getattr(random, name)
    return makeNamespace(math, drawbot, **additionalNames)


def runScript(sourcePath, namespace=None):
    with open(sourcePath, encoding="utf-8") as f:
        source = f.read()
    runScriptSource(source, sourcePath, namespace)


def runScriptSource(source, sourcePath, namespace=None):
    if namespace is None:
        namespace = {}
    if os.path.exists(sourcePath):
        sourcePath = os.path.normpath(os.path.abspath(sourcePath))
    namespace.update({"__name__": "__main__", "__file__": sourcePath})

    if sourcePath:
        sourcePath = os.path.normpath(os.path.abspath(sourcePath))
        parentDir = os.path.dirname(sourcePath)
        saveDir = os.getcwd()
    else:
        parentDir = None
        saveDir = None

    code = compile(source, sourcePath, "exec")

    if parentDir is not None:
        os.chdir(parentDir)
    try:
        exec(code, namespace)
    finally:
        if saveDir is not None:
            os.chdir(saveDir)



from fontTools.misc.transform import Transform


size(200, 200)


with savedState():
    rotate(15, center=(50, 50))
    rect(25, 25, 50, 50)


with savedState():
    scale(1.2, center=(150, 50))
    rect(125, 25, 50, 50)


with savedState():
    skew(15, 10, center=(50, 150))
    rect(25, 125, 50, 50)


with savedState():
    t = Transform()
    t = t.scale(0.9, 1.1)
    t = t.rotate(0.2)
    transform(t, center=(150, 150))
    rect(125, 125, 50, 50)



size(200, 200)
fill(1)
rect(0, 0, 200, 200)
fill(0)

path = BezierPath()
path.rect(0, 0, 100, 100)
path.scale(0.9, center=(50, 50))
path.rotate(30, center=(50, 50))
path.skew(15, 0, center=(50, 50))
path.translate(20, 20)
drawPath(path)



size(200, 200)

path = BezierPath()
path.moveTo((20, 20))
path.arcTo((20, 180), (180, 180), 30)
path.arcTo((180, 180), (110, 20), 30)
path.lineTo((110, 20))

stroke(0)
strokeWidth(8)
fill(None)

drawPath(path)



from types import SimpleNamespace
import functools
import uharfbuzz as hb
from .font import intToTag, tagToInt


def getShapeFuncForSkiaTypeface(skTypeface):
    face = makeHBFaceFromSkiaTypeface(skTypeface)
    font = hb.Font(face)
    return functools.partial(_shape, face, font)


def makeHBFaceFromSkiaTypeface(skTypeface):
    tableData = {}
    tableTags = {intToTag(tag) for tag in skTypeface.getTableTags()}

    def getTable(face, tag, userData):
        if tag in tableData:
            return tableData[tag]
        if tag not in tableTags:
            return None
        data = skTypeface.getTableData(tagToInt(tag))
        # HB doesn't hold on the data, and neither does Skia, so we
        # need to do that ourselves.
        tableData[tag] = data
        return data

    return hb.Face.create_for_tables(getTable, None)


def _shape(face, font,
           text, fontSize=None,
           startPos=(0, 0), startCluster=0,
           flippedCanvas=False,
           *,
           features=None,
           variations=None,
           direction=None,
           language=None,
           script=None):
    if features is None:
        features = {}
    if variations is None:
        variations = {}

    if fontSize is None:
        fontScaleX = fontScaleY = 1
    else:
        fontScaleX = fontScaleY = fontSize / face.upem
    if flippedCanvas:
        fontScaleY = -fontScaleY

    font.scale = (face.upem, face.upem)
    font.set_variations(variations)

    hb.ot_font_set_funcs(font)

    buf = hb.Buffer.create()
    buf.add_str(text)  # add_str() does not accept str subclasses
    buf.guess_segment_properties()
    buf.cluster_level = hb.BufferClusterLevel.MONOTONE_CHARACTERS

    if direction is not None:
        buf.direction = direction
    if language is not None:
        buf.language = language
    if script is not None:
        buf.script = script

    hb.shape(font, buf, features)

    gids = [info.codepoint for info in buf.glyph_infos]
    clusters = [info.cluster + startCluster for info in buf.glyph_infos]
    positions = []
    startPosX, startPosY = startPos
    x = y = 0
    for pos in buf.glyph_positions:
        dx, dy, ax, ay = pos.position
        positions.append((
            startPosX + (x + dx) * fontScaleX,
            startPosY + (y + dy) * fontScaleY,
        ))
        x += ax
        y += ay
    endPos = startPosX + x * fontScaleX, startPosY + y * fontScaleY
    return SimpleNamespace(
        gids=gids,
        clusters=clusters,
        positions=positions,
        endPos=endPos,
    )


def scalePositions(positions, sx, sy=None):
    if sy is None:
        sy = sx
    return [(x * sx, y * sy) for x, y in positions]


def getFeatures(face, otTableTag):
    features = set()
    for scriptIndex, script in enumerate(hb.ot_layout_table_get_script_tags(face, otTableTag)):
        langIdices = list(range(len(hb.ot_layout_script_get_language_tags(face, otTableTag, scriptIndex))))
        langIdices.append(0xFFFF)
        for langIndex in langIdices:
            features.update(hb.ot_layout_language_get_feature_tags(face, otTableTag, scriptIndex, langIndex))
    return sorted(features)



import os
import subprocess
import sys


# Monkeypatch this variable to override the ffmpeg executable location
FFMPEG_PATH = None


def generateMP4(
        imageTemplate,
        mp4path,
        frameRate,
        codec="libx264",
        preferPyFFmpeg=False):
    ffmpegPath = FFMPEG_PATH
    if ffmpegPath is None and not preferPyFFmpeg:
        ffmpegPath = findExecutable("ffmpeg")
    if ffmpegPath is None:
        ffmpegPath = getPyFFmpegPath()
    cmds = [
        ffmpegPath,             # path to the ffmpeg executable
        "-y",                   # overwrite existing files
        "-loglevel", "16",      # 'error, 16' Show all errors, including ones which can be recovered from.
        "-r", str(frameRate),   # frame rate
        "-i", imageTemplate,    # input sequence
        "-c:v", codec,          # codec
        "-crf", "20",           # Constant Rate Factor
        "-pix_fmt", "yuv420p10le",  # pixel format. 8bit: yuv420p 10bit: yuv420p10le, yuv422p10le, yuv444p10le
        mp4path,                # output path
    ]
    runExternalProcess(cmds)


def getPyFFmpegPath():
    try:
        import pyffmpeg
    except ImportError:
        raise ImportError("ffmpeg not found: install ffmpeg manually, or do 'pip install pyffmpeg==1.6.1'")
    ffmpegPath = os.path.normpath(pyffmpeg.FFmpeg().get_ffmpeg_bin())
    st = os.stat(ffmpegPath)
    if sys.platform != "win32" and not st.st_mode & 0o00100:
        # https://github.com/deuteronomy-works/pyffmpeg/issues/33
        os.chmod(ffmpegPath, st.st_mode | 0o00100)
    return ffmpegPath


def findExecutable(name):
    """
        >>> import sys
        >>> findExecutable("python") == sys.executable
        True
        >>> findExecutable("ls")
        '/bin/ls'
    """
    if sys.platform == "win32":
        name += ".exe"
    path = os.getenv("PATH")
    if path:
        path = path.split(os.pathsep)
        if sys.platform != "win32" and "/usr/local/bin" not in path:
            path.append("/usr/local/bin")
        for p in path:
            ep = os.path.join(p, name)
            if os.path.exists(ep):
                return ep
    return None


def runExternalProcess(cmds, cwd=None):
    r"""
        >>> stdout, stderr = runExternalProcess(["which", "ls"])
        >>> stdout
        '/bin/ls\n'
        >>> assert stdout == '/bin/ls\n'
        >>> runExternalProcess(["which", "fooooo"])
        Traceback (most recent call last):
            ...
        subprocess.CalledProcessError: Command '['which', 'fooooo']' returned non-zero exit status 1.
        >>> stdout, stderr = runExternalProcess(["python", "-S", "-c", "print('hello')"])
        >>> stdout
        'hello\n'
    """
    p = subprocess.Popen(
        cmds,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        universal_newlines=True,
    )
    stdoutdata, stderrdata = p.communicate()
    assert p.returncode is not None
    if p.returncode != 0:
        sys.stdout.write(stdoutdata)
        sys.stderr.write(stderrdata)
        raise subprocess.CalledProcessError(p.returncode, cmds)
    return stdoutdata, stderrdata


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import struct
from fontTools.ttLib import TTFont


def makeTTFontFromSkiaTypeface(skTypeface):
    ttf = TTFont(lazy=True)
    ttf.reader = SkiaSFNTReader(skTypeface)
    ttf._tableCache = None  # fonttools fix: doesn't get set when not reading from a file
    return ttf


class SkiaSFNTReader:

    def __init__(self, skTypeface):
        self.skTypeface = skTypeface
        self.tags = {intToTag(tagInt): tagInt for tagInt in skTypeface.getTableTags()}

    def __len__(self):
        return len(self.tags)

    def __contains__(self, tag):
        return tag in self.tags

    def keys(self):
        return self.tags.keys()

    def __getitem__(self, tag):
        if tag not in self.tags:
            raise KeyError(tag)
        return self.skTypeface.getTableData(self.tags[tag])


def intToTag(intTag):
    return struct.pack(">i", intTag).decode()


def tagToInt(tag):
    if isinstance(tag, str):
        tag = bytes(tag, "ascii")
    return struct.unpack(">i", tag)[0]



size(200, 200)

path = BezierPath()
path.oval(5, 5, 190, 190)

fill(1, 0.3, 0)
drawPath(path)

fill(0)

step = 10
r = 2
d = r * 2
for x in range(0, width() + step, step):
    for y in range(0, height() + step, step):
        if path.pointInside((x, y)):
            oval(x - r, y - r, d, d)
